
/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*             Damien Doligez, projet Para, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: scions_gc.c,v 1.1 1999/09/30 13:06:30 lefessan Exp $ */

#include "compact.h"
#include "config.h"
#include "fail.h"
#include "freelist.h"
#include "gc.h"
#include "gc_ctrl.h"
#include "major_gc.h"
#include "misc.h"
#include "mlvalues.h"
#include "roots.h"
#include "weak.h"
#include "join.h"
#include "scions_gc.h"
#include <stdio.h>

#ifdef __STDC__
#include <limits.h>
#else
#ifdef ARCH_SIXTYFOUR
#define LONG_MAX 0x7FFFFFFFFFFFFFFF
#else
#define LONG_MAX 0x7FFFFFFF
#endif
#endif

value *scions_vals, *scions_vals_cur, *scions_vals_end, *scions_vals_ptr;
int scions_vals_size;
value current_stamp;
value *scions_gray_vals, *scions_gray_vals_cur, *scions_gray_vals_end;
int scions_gray_vals_size;

value root_distance = Val_int(6);

void init_scion_gc()
{
  scions_vals_size = 128;
  scions_vals = (value *) malloc (scions_vals_size * sizeof (value));
  if (scions_vals == NULL)
    fatal_error ("Fatal error: not enough memory for the initial scions.\n");
  scions_vals_cur = scions_vals;
  scions_vals_end = scions_vals + scions_vals_size;
  
  scions_gray_vals_size = 128;
  scions_gray_vals = (value *) malloc (scions_gray_vals_size * sizeof (value));
  if (scions_gray_vals == NULL)
    fatal_error ("Fatal error: not enough memory for the initial scions.\n");
  scions_gray_vals_cur = scions_gray_vals;
  scions_gray_vals_end = scions_gray_vals + scions_gray_vals_size;
  
}

value set_root_distance(value new)
{
  root_distance = new;
  return Val_unit;
}


/* Cyclic Garbage Collection */

void realloc_scions_vals (void)
{
  value *new;

  Assert (scions_vals_cur == scions_vals_end);
  gc_message (0x08,"Growing scions_vals to %luk bytes\n",
	      (long) scions_vals_size * sizeof (value) / 512);
  new = (value *) realloc ((char *) scions_vals,
                             2 * scions_vals_size * sizeof (value));
  if (new == NULL){
    fatal_error ("No room for growing scions_vals\n");    
  }else{
    scions_vals = new;
    scions_vals_cur = scions_vals + scions_vals_size;
    scions_vals_size *= 2;
    scions_vals_end = scions_vals + scions_vals_size;
  }
}

void realloc_scions_gray_vals (void)
{
  value *new;

  Assert (scions_gray_vals_cur == scions_gray_vals_end);
  gc_message (0x08,"Growing scions_gray_vals to %luk bytes\n",
	      (long) scions_gray_vals_size * sizeof (value) / 512);
  new = (value *) realloc ((char *) scions_gray_vals,
                             2 * scions_gray_vals_size * sizeof (value));
  if (new == NULL){
    fatal_error ("No room for growing scions_gray_vals\n");    
  }else{
    scions_gray_vals = new;
    scions_gray_vals_cur = scions_gray_vals + scions_gray_vals_size;
    scions_gray_vals_size *= 2;
    scions_gray_vals_end = scions_gray_vals + scions_gray_vals_size;
  }
}


/* define a specific function instead of "compare" since root marks 
   smaller than generated marks for "compare", but not for the DGC */

extern int compare(value,value);
int compare_stamps(value left_stamp,value right_stamp)
{
  value left_mark = Field(left_stamp,STAMP_MARK);
  value right_mark = Field(right_stamp,STAMP_MARK);
  int retval;

  if(left_stamp == Val_unit){
    if(right_stamp == Val_unit){
      value left_d = Field(left_stamp,STAMP_DISTANCE);
      value right_d = Field(right_stamp,STAMP_DISTANCE);

      /* compare two root marks by distances */
      if(left_d == right_d) return 0;
      if(left_d < right_d) return -1;
      return 1;
    }
    /* left is root mark, right is not */
    return 1;
  } else
    if(right_stamp == Val_unit){
    /* right is root mark, left is not */
      return -1;
    }

  /* compare two generated marks */
  retval = compare(left_mark,right_mark);
  if(retval == 0){
    value left_c = Field(left_stamp,STAMP_COLOR);    
    value right_c = Field(right_stamp,STAMP_COLOR);
    
    if(left_c == right_c){
      value left_d = Field(left_stamp,STAMP_DISTANCE);
      value right_d = Field(right_stamp,STAMP_DISTANCE);
      
      if(left_d == right_d) return 0;
      if(left_d < right_d) return -1;
      return 1;
    }
    if(left_c < right_c) return -1;
    return 1;
  }
  return retval;
}

value compare_stamps_val(value left,value right)
{
  return Val_int(compare_stamps(left,right));
}

int compare_scions(value left,value right)
{
  return compare_stamps(Field(left,GCSCION_NEXT_STAMP),
			Field(right,GCSCION_NEXT_STAMP));
}


static int (*compar)(value,value);
static value *tab;

static void qsort_rec(int left,int right)
{
  value left_c = left;
  value right_c = right-1;
  value pivot = tab[left_c];
  value test;
  
go_down:
  test = tab[right_c];
  if(compar(pivot,test) >=0 ){
    tab[left_c] = test;
    goto go_left;
  }
  
go_right:
  right_c--;
  if(left_c == right_c) goto go_end;
  goto go_down;
  
go_up:
  test = tab[left_c];
  if(compar(pivot,test) <0 ){
    tab[right_c] = test;
    goto go_right;
  }
  
go_left:
  left_c++;
  if(left_c == right_c) goto go_end;
  goto go_up;
  
go_end:
  tab[left_c] = pivot;

  right_c = left_c - left;
  if(right_c < 3){
    value tmp1 = tab[left];
    value tmp2 = tab[left+1];
    
    if( right_c == 2 && compar(tmp1,tmp2) > 0){
      tab[left] = tmp2;
      tab[left+1] = tmp1;
    }
  } else
    qsort_rec(left,left_c);
  
  right_c = right - left_c;
  if(right_c < 4){
    value tmp1 = tab[left_c+1];
    value tmp2 = tab[left_c+2];
    
    if( right_c == 3 && compar(tmp1,tmp2) > 0){
      tab[left_c+1] = tmp2;
      tab[left_c+2] = tmp1;
    }
  } else
    qsort_rec(left_c+1,right);
}


static void my_qsort(int left,int right)
{
  int d = right - left;
  
  if( d <= 2) {
    
    if( d == 2){
      value tmp1 = scions_vals[left];
      value tmp2 = scions_vals[left+1];

      if(compare_scions(tmp1,tmp2) > 0){
	scions_vals[left] = tmp2;
	scions_vals[left+1] = tmp1;
      }
    }
    return;
  }
  tab = scions_vals; /* specialized */
  compar = compare_scions; /* specialized */
  qsort_rec(left,right);
}


/* prevent old values from being traced:
     These old values should not be added to gray_vals, since they should be
     marked with another mark than rootmark.
 */
value delete_bck_stub(value gcscion)
{
  Field(gcscion,GCSCION_BCK_STUB) = Val_unit;
  return Val_unit;
}

value delete_gcscion(value gcscion)
{
  value next;

 next:
  Field(gcscion,GCSCION_BCK_STUB) = Val_unit;
  Field(gcscion,GCSCION_TARGET) = Val_unit;
  Field(gcscion,GCSCION_GCSTUB) = Val_unit;
  next = Field(gcscion,GCSCION_NEXT);
  if( next== Val_unit)
    return Val_unit;

  Field(gcscion,GCSCION_NEXT) = Val_unit;
  gcscion = next;
  goto next;
}

#define DarkenChild(child,vals)                                   \
   if (Is_block (child) && Is_in_heap (child)) {                  \
	    hd = Hd_val(child);                                   \
	    if (Tag_hd(hd) == Infix_tag) {                        \
	      child -= Infix_offset_val(child);                   \
	      hd = Hd_val(child);                                 \
	    }                                                     \
	    if (Is_white_hd (hd)){                                \
	      Hd_val (child) = Grayhd_hd (hd);                    \
	      *vals##_ptr++ = child;                              \
	      if (vals##_ptr >= vals##_end) {                     \
		vals##_cur = vals##_ptr;                          \
		realloc_##vals ();                                \
		vals##_ptr = vals##_cur;                          \
    }}}


void mark_scions_slice (long int work)
{
  value *scions_gray_vals_ptr;  /* Local copy of scions_gray_vals_cur */
  value v, child;
  header_t hd;
  mlsize_t size, i;
  value min_stamp,stamp;

  if( gc_phase == TRACE_LEFTROOTS ){
    gc_phase = TRACE_LEFTSCIONS;
    my_qsort(0,scions_vals_cur-scions_vals);
    scions_vals_ptr = scions_vals;
    while(scions_vals_ptr < scions_vals_cur){
      v = *scions_vals_ptr++;
      /* don't darken old value of Field(v,GCSCION_STAMP) since it can
      not appear in the CAML stack */
      Field(v,GCSCION_STAMP) = Field(v,GCSCION_NEXT_STAMP);
    }
    scions_vals_ptr = scions_vals - 1;
    scions_gray_vals_cur = scions_gray_vals;
  } else
    if( gc_phase == TRACE_RIGHTROOTS ){
      gc_phase = TRACE_RIGHTSCIONS;
      scions_vals_ptr = scions_vals_cur;
      scions_gray_vals_cur = scions_gray_vals;
    }
  
  scions_gray_vals_ptr = scions_gray_vals_cur;
  while (work > 0){
    if (scions_gray_vals_ptr > scions_gray_vals){
      v = *--scions_gray_vals_ptr;
      DPRINT1("Scanning v = %lx \n",v);
      hd = Hd_val(v);
      Assert (Is_gray_hd (hd));
      Hd_val (v) = Blackhd_hd (hd);
      size = Wosize_hd(hd);
      if (Tag_hd (hd) < No_scan_tag){
	if(Tag_hd(hd) == GcStub_tag){
	  gc_message(0x08,"!",0);
	  /* what about stubs ? */
	  if(gc_phase & LEFT_FLAG){
	    stamp = Field(v,GCSTUB_MIN);
	    child = Field(stamp,STAMP_MARK);
	    DarkenChild(child,scions_gray_vals);
	    Field(stamp,STAMP_MARK) = Field(current_stamp,STAMP_MARK);
	    Field(stamp,STAMP_COLOR) = Field(current_stamp,STAMP_COLOR);
	    Field(stamp,STAMP_DISTANCE) = Field(current_stamp,STAMP_DISTANCE);
	  } else {
	    min_stamp = Field(v,GCSTUB_MIN);
	    stamp = Field(v,GCSTUB_MAX);
	    child = Field(stamp,STAMP_MARK);
	    DarkenChild(child,scions_gray_vals);
	    if(compare_stamps(min_stamp,current_stamp) > 0){
	      Field(stamp,STAMP_MARK) = Field(min_stamp,STAMP_MARK);
	      Field(stamp,STAMP_COLOR) = Field(min_stamp,STAMP_COLOR);
	      Field(stamp,STAMP_DISTANCE) = Field(min_stamp,STAMP_DISTANCE);
	    } else {
	      int d = compare(Field(current_stamp,STAMP_MARK),
			      Field(min_stamp,STAMP_MARK));
	      if(Field(current_stamp,STAMP_COLOR) == COLOR_WHITE &&
		 Field(min_stamp,STAMP_COLOR) == COLOR_WHITE &&
		 d == 0){
		Field(stamp,STAMP_MARK) = Field(min_stamp,STAMP_MARK);
		Field(stamp,STAMP_COLOR) = COLOR_WHITE;
		Field(stamp,STAMP_DISTANCE) = Field(current_stamp,STAMP_DISTANCE);
	      } else {
		Field(stamp,STAMP_MARK) = Field(current_stamp,STAMP_MARK);
		Field(stamp,STAMP_COLOR) = COLOR_GRAY;
		Field(stamp,STAMP_DISTANCE) = Field(current_stamp,STAMP_DISTANCE);		
		if(d == 0 && 
		   Field(stamp,STAMP_DISTANCE) < Field(min_stamp,STAMP_DISTANCE)){
		  Field(stamp,STAMP_DISTANCE) = Field(min_stamp,STAMP_DISTANCE);
		  
		}

		if(Field(current_stamp,STAMP_COLOR) == COLOR_GRAY &&
		   Field(v,GCSTUB_BCK_MARK) != Val_unit)
		  {	
		    /* optimistic back-tracing */
		    value bck_mark = Field(Field(v,GCSTUB_BCK_MARK),0);
		    value bck_trace = Field(v,GCSTUB_BCK_TRACE);
		    value *scion_ptr = scions_vals_ptr;
		    value scion;

		    DarkenChild(bck_trace,scions_gray_vals);
		  next_scion:
		    scion = *scion_ptr;
		    stamp = Field(scion,GCSCION_STAMP);
		    if(Field(stamp,STAMP_COLOR) == COLOR_GRAY &&
		       compare(Field(stamp,STAMP_MARK),bck_mark) == 0
		       ) {
		      if(Field(scion,GCSCION_BCK_MARK) == Val_unit &&
			 Field(scion,GCSCION_BCK_STUB) == Val_unit &&
			 Field(scion,GCSCION_NEXT) == Val_unit){
			Field(scion,GCSCION_BCK_MARK) = 
			  Field(v,GCSTUB_BCK_MARK);
			Field(scion,GCSCION_BCK_STUB) = v;
			Field(scion,GCSCION_NEXT) = bck_trace;
			bck_trace = scion;
		      }
		      if(scion_ptr > scions_vals){
			--scion_ptr;
			goto next_scion;
		      }
		    }	
		    Field(v,GCSTUB_BCK_TRACE) = bck_trace;
		  }
	      }
	    }
	  } 
	}
	DPRINT1("size = %d \n",size);
	for (i = 0; i < size; i++){
	  child = Field (v, i);
	  DarkenChild(child,scions_gray_vals);
        }
      }
      work -= Whsize_wosize(size);
    }else {

      if((scions_vals_ptr == scions_vals && gc_phase & RIGHT_FLAG) ||
	 (scions_vals_ptr+1 == scions_vals_cur && gc_phase & LEFT_FLAG)) {
	
	/* Marking is really done. */
	update_weak_pointers ();
	
	/* Initialise the sweep phase. */
	gc_sweep_hp = heap_start;
	fl_init_merge ();
	if(gc_phase & LEFT_FLAG){
	  gc_phase = TRACE_SWEEPLEFT;
	} else {
	  gc_phase = TRACE_SWEEPRIGHT;
	}
	chunk = heap_start;
	gc_sweep_hp = chunk;
	limit = chunk + Chunk_size (chunk);
	work = 0;
      } else {

	/* pass to next scion */
	if( gc_phase & RIGHT_FLAG)
	  scions_vals_ptr--;
	else 
	  scions_vals_ptr++;
	v = *scions_vals_ptr;
        current_stamp = Field(v, GCSCION_STAMP);
	hd = Hd_val(v);
	size = Wosize_hd(hd);
	for (i = 0; i < size; i++){
	  child = Field (v, i);
	  DarkenChild(child,scions_gray_vals);
	}
      }
    }
  }
  scions_gray_vals_cur = scions_gray_vals_ptr;
}
