(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Obj
open Printf
open Buffer
open Text
open Efuns
  
let location = {
    loc_global_map = Keymap.create ();
    loc_windows = [];
    loc_buffers = Hashtbl.create 13;
    loc_files = Hashtbl.create 13;
    loc_dirname = Sys.getcwd ();
    loc_width = !width;
    loc_height = !height;
    loc_fg = !fg;
    loc_bg = !bg;
    loc_font = !font;
    loc_local_vars = Vars.empty;
    loc_counter = 0;
    
    loc_vars_table = Efuns.vars_table;
    loc_fonts = Hashtbl.create 37;
    loc_fonts_names = Array.create 256 "";
    loc_fonts_n = 0;
    loc_colors = Hashtbl.create 37;
    loc_colors_names = Array.create 256 "";
    loc_colors_n = 0;
    
    loc_mutex = Concur.Mutex.create ()
} 

exception SigInt
let _ =
  Sys.set_signal Sys.sigint (Sys.Signal_handle (fun _ ->  
        (* Should auto-save all buffers, and then exit ... *)
        exit 1)); 
  Sys.set_signal Sys.sigterm (Sys.Signal_handle 
      (fun _ ->  
        (* Should auto-save all buffers, and then exit ... *)
        raise SigInt));
  Sys.set_signal Sys.sighup (Sys.Signal_handle (fun _ ->  raise SigInt))
  
let _ =
(* color 0 is foreground *)
  Window.get_color location !fg;
(* color 1 is background *)
  Window.get_color location !bg;
(* color 2 is highlight *)
  Window.get_color location "cyan";
(* font 0 is initial font *)
  Window.get_font location !font

let _ =
  Efuns.init location (* launch first hooks *)
  
  
let display = Xterm.create_display !displayname 
  location.loc_colors_names location.loc_fonts_names "Efuns" "efuns"
let top_window = Top_window.create location display

  
  
  
  
  
  
open Select
  
let offers = ref []
  
let register_register () =
      let def register_me (local, name, args) =
      reply |
        { let name = match args with
            | [name] -> name
            | _ -> ":0"
          in
          Printf.printf "Display <%s>" name; print_newline ();
          offers := (name, local) :: !offers; }

  in      
  offers := [];
  print_string "-----------------------"; print_newline ();
  Ns.register !ns_name (register_me: (Join.location * string * string list -> unit)) vartype
      
let restart location dpyname = 
  
  Concur.Mutex.unlock location.loc_mutex;
  
  let display = Xterm.create_display !displayname 
      location.loc_colors_names location.loc_fonts_names "Efuns" "efuns"
  in
  display.Xterm.dpy_highlighted <- 2;
  List.iter (fun top ->
      let xterm = Xterm.create_window 
          display "new_frame"
          location.loc_font 
          top.top_width
          top.top_height in
      top.top_display <- Some  display;
      top.top_xterm <- Some xterm;
      Xterm.install_handler display xterm 
        (Top_window.handler top xterm);      
  ) location.loc_windows;
  
  Top_window.update_display location; 
  
  register_register ();
  
  (* Main loop *)
  let rec loop () =
    try
      Xterm.event_loop ()
    with
      SigInt -> loop ()
  in
  loop ()

let goto location dpyname destination =
      (* clean up all dependencies towards X components *)
  List.iter (fun top ->
      match top.top_display with
        None -> ()
      | Some dpy ->
          Xterm.close_display dpy;
          top.top_display <- None;
          top.top_xterm <- None;
          Window.iter (fun frame -> frame.frm_redraw <- true) top.top_windows;
  ) location.loc_windows;
  offers := [];
  let loc agent 
        [
            (* Threads *)
        Concur;
            
            (* Xlib modules *)
        Xdebug; Xrm; Xfifo; Jeton; Xtypes; Xconst; Xbuffer; Xauth;
        Ll_trans; Conv_event; Display; Xproto; X; Xsync; Xasync; 
        XA; XC; XK; KeyBind; Xlib; Xutils; Icccm; Zpixmap; Xterm;

            (* Str modules *)
        Str;
            
            (* Dynlink modules *)
        Dynlink;
            
            (* Efuns modules *)
        Str2; Version; Time; Lexers; Parsers; Misc; Efuns; Text;  
        Keymap; Buffer; Window; Frame; Top_window; Simple; Minibuffer;
        Multi_frames; Select; Search; System; Abbrevs; Complex; 
        Compile; Eval; Interactive; Accents_mode; Ocamlc; Printtyp;
        Ocaml_mode;  Makefile_mode; Tex_mode; C_mode; Lisp_mode;
        Html_mode; Std_efunsrc
      ] def
        agent_goto (destination,location,dpyname) = 
        { go destination; { reply () | restart location dpyname; }}
  in        
  agent_goto (destination,location,dpyname);
  exit 0
  
 
let migrate frame =
  let location = frame.frm_location in
  match !offers with
    [] -> failwith "No destination available"
  | [dpyname,destination] -> goto location dpyname destination
  | _ ->
      simple_select frame "Go to display :"
        (fun dpyname ->
          let destination = List.assoc dpyname !offers in
         goto location dpyname destination)
 
let _ = (* register our service in the name server *)
  Keymap.add_interactive  (Location location) "migrate" migrate;
  Ns.name := !ns_host;
  Ns.port := !ns_port;
  register_register ()
  
  
  
  
  
  
  
let _ =
  display.Xterm.dpy_highlighted <- 2;
  Dynlink.init ();
  Dynlink.allow_unsafe_modules true;
  Eval.load top_window "Efunsrc";
  Efuns.init location; (* launch second hooks *)
  Interactive.create_bindings location;

(* open the fisrt buffers *)
  List.iter (fun name ->
    Frame.load_file top_window.top_windows name) 
    !init_files;
  Top_window.update_display location;

  (* Main loop *)
  let rec loop () =
    try
      Xterm.event_loop ()
    with
      SigInt -> loop ()
  in
  loop ()

  



  
