(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)


open Interactive
open Efuns
open Keymap
open Simple
open Select
open Search
open Eval
open Compile
open Abbrevs
open Complex
open System
open Multi_frames
  
let _ =
  dont_complete := !dont_complete @ 
    [".*\.cm.";".*\.cmxa";".*~";".*\.a";"core";"\..*"]
  
let init_global_map location = 
  add_interactive (Location location) "replace-string" replace_string;
  add_interactive (Location location) "replace-regexp" replace_regexp;
  add_interactive (Location location) "query-replace-string" query_replace_string;
  add_interactive (Location location) "query-replace-regexp" query_replace_regexp;
  add_interactive (Location location) "load-library" load_library;
  add_interactive (Location location) "get-position" get_pos;
  add_interactive (Location location) "open-display" open_display;
  add_interactive (Location location) "accents-mode" 
    (fun frame -> Accents_mode.install frame.frm_buffer);
  add_interactive (Location location) "compile" (compile c_find_error);
  add_interactive (Location location) "goto-line" goto_line;
  add_interactive (Location location) "goto-char" goto_char;
  add_interactive (Location location) "unset-attr" unset_attr;
  
  let gmap = location.loc_global_map in
  let c_h = (ControlMap, Char.code 'h') in
  let c_x = (ControlMap, Char.code 'x') in
  let c_c = (ControlMap, Char.code 'c') in
  let n_5 = (NormalMap, Char.code '5') in

(* standard keys *)
  interactive (Location location) [NormalMap, XK.xk_BackSpace] "delete-before-point" 
    delete_backspace_char;
  interactive (Location location) [NormalMap, XK.xk_Delete] "delete-at-point" delete_char;
  interactive (Location location) [NormalMap, XK.xk_Return] "insert-return" insert_return;
  interactive (Location location) [NormalMap, XK.xk_Left] "move-backward" 
  (fun frame -> move_backward frame 1);
  interactive (Location location) [NormalMap, XK.xk_Right] "move-forward" 
  (fun frame -> move_forward frame 1);
  interactive (Location location) [NormalMap, XK.xk_Down] "forward-line" forward_line;  
  interactive (Location location) [NormalMap, XK.xk_Up] "backward-line" backward_line;
  interactive (Location location) [NormalMap, XK.xk_Prior] "backward-screen" backward_screen;
  interactive (Location location) [NormalMap, XK.xk_Next] "forward-screen" forward_screen;


(* Mouse *)
  interactive (Location location) [NormalMap, XK.xk_Pointer_Button1]
    "set-active-frame" mouse_set_frame;
  interactive (Location location) [NormalMap, XK.xk_Pointer_Button2]
    "insert-at-point" mouse_yank_at_click;
  interactive (Location location) [NormalMap, XK.xk_Pointer_Button3]
    "insert-at-point" mouse_save_then_kill;
    
(* C-  map *)
  interactive (Location location) [ControlMap, XK.xk_Up] "backward-paragraph" (to_frame backward_paragraph);
  interactive (Location location) [ControlMap, Char.code 'w'] "kill-region" kill_region;
  interactive (Location location) [ControlMap, XK.xk_Down] "forward-paragraph" (to_frame forward_paragraph);  
  interactive (Location location) [ControlMap, Char.code 'a']  
    "beginning-of-line" beginning_of_line;
  interactive (Location location) [ControlMap, Char.code 'd']  "delete-at-point" delete_char;
  interactive (Location location) [ControlMap, Char.code 'b']
    "move-backward" move_backward;
  interactive (Location location) [ControlMap, Char.code 'f']  "move-forward" move_forward;
  interactive (Location location) [ControlMap, Char.code 'e'] "end-of-line" end_of_line;
  interactive (Location location) [ControlMap, Char.code 'k']
    "kill_end-of-line" kill_end_of_line;
  interactive (Location location) [ControlMap, Char.code 's'] 
    "isearch-forward" isearch_forward    ;
  interactive (Location location) [ControlMap, Char.code 'r'] 
    "isearch-backward" isearch_backward    ;
  interactive (Location location) [ControlMap, Char.code '_'] 
    "undo" undo;
  interactive (Location location) [ControlMap, XK.xk_Right ] "forward-word" (to_frame forward_word);  
  interactive (Location location) [ControlMap, XK.xk_Left ] "backward-word" (to_frame backward_word);
  interactive (Location location) [ControlMap, XK.xk_Next] "end-of-file" end_of_file;
  interactive (Location location) [ControlMap, XK.xk_Prior] "begin-of-file" begin_of_file;
  interactive (Location location) [ControlMap, Char.code 'y'] "insert-killed" insert_killed;
  interactive (Location location) [ControlMap, Char.code ' '] "mark-at-point" mark_at_point;
  interactive (Location location) [ControlMap, Char.code 'l'] "recenter" recenter;
  interactive (Location location) [ControlMap, Char.code 't'] "transpose-chars" (to_frame transpose_chars);
  interactive (Location location) [ControlMap, XK.xk_BackSpace] "hungry-electric-delete" hungry_electric_delete;
  
(* M- map *)
  interactive (Location location) [MetaMap, Char.code 's'] 
    "isearch-forward-regexp" isearch_forward_regexp;
  interactive (Location location) [MetaMap, Char.code 'r'] 
    "isearch-backward-regexp" isearch_backward_regexp;
  interactive (Location location) [MetaMap, XK.xk_Right ] "forward-word" (to_frame forward_word)  ;
  interactive (Location location) [MetaMap, XK.xk_Left ] "backward-word" (to_frame backward_word);
  interactive (Location location) [MetaMap, Char.code 'd' ]
    "delete-forward-word" (to_frame delete_forward_word);
  interactive (Location location) [MetaMap, XK.xk_BackSpace ] 
    "delete-backward-word" (to_frame delete_backward_word);
  interactive (Location location) [MetaMap, Char.code 'x'] "call-interactive" call_interactive;
  interactive (Location location) [MetaMap, Char.code '%']
    "query-replace-string" query_replace_string;
  interactive (Location location) [MetaMap, Char.code '!']
    "shell-command" shell_command;
  interactive (Location location) [MetaMap, Char.code 'y']
    "insert-next-killed" insert_next_killed;
  interactive (Location location) [MetaMap, Char.code 't'] "transpose-words" 
  (to_frame transpose_words);

(* C-x map *)
  Keymap.add_prefix gmap [c_x];
  interactive (Location location) [c_x; ControlMap, Char.code 'f']
    "load-buffer" load_buffer;
  interactive (Location location) [c_x; NormalMap, Char.code 'i']
    "insert-file" insert_file;
  interactive (Location location) [c_x; ControlMap, Char.code 'c']
    "exit" exit_efuns; 
  interactive (Location location) [c_x; NormalMap, Char.code 'b']
    "change-buffer" change_buffer;
  interactive (Location location) [c_x; NormalMap, Char.code 'F']
    "change-font" change_font;
  interactive (Location location) [c_x; NormalMap, Char.code '2']
    "vertical-cut-frame" v_cut_frame;    
  interactive (Location location) [c_x; NormalMap, Char.code '3']
    "horizontal-cut-frame" h_cut_frame;    
  interactive (Location location) [c_x; NormalMap, Char.code '1']
    "one-frame" one_frame;
  interactive (Location location) [c_x; NormalMap, Char.code '0']
    "delete-frame" delete_frame;
  interactive (Location location) [c_x; NormalMap, Char.code 'o']
    "next-frame" next_frame;
  interactive (Location location) [c_x; NormalMap, Char.code 'k']
    "kill-buffer" kill_buffer;
  interactive (Location location) [c_x; ControlMap, Char.code 's']
    "save-buffer" save_buffer; 
  interactive (Location location) [c_x;NormalMap, Char.code '`' ] "next-error" next_error;
  interactive (Location location) [c_x;ControlMap, Char.code 'w'] "write-file" write_buffer;
  
  interactive (Location location) [c_x; NormalMap, Char.code 's'] "save-some-buffers" save_some_buffers;
  interactive (Location location) [c_x; ControlMap, Char.code 'x'] "point-at-mark" point_at_mark;
  
(* C-x 5 map *)
  Keymap.add_prefix gmap [c_x; n_5];
  interactive (Location location) [c_x; n_5; NormalMap, Char.code 'f']
    "window-load-buffer" window_load_buffer;
  interactive (Location location) [c_x; n_5; NormalMap, Char.code 'b']
    "window-change-buffer" window_change_buffer;
  interactive (Location location) [c_x; n_5; NormalMap, Char.code '0']
    "delete-window" Top_window.delete_window;

(* C-h map *)
  Keymap.add_prefix gmap [c_h];
  interactive (Location location) [c_h; NormalMap, Char.code 'K'] 
    "help-bindings" Frame.bindings_help
  
let _ =
  Efuns.add_start_hook init_global_map
