open Unix
open ThreadUnix
  open Jtypes
open Jprims
open Space

external compare_stamps : stamp -> stamp -> int = "compare_stamps_val"

(* gcscion.bck_stub should never be modified nor access in CAML, since then 
the stub would be marked with rootmark, breaking any hope to detect the cycle.
*)
external delete_bck_stub' : gcscion_option_pointer -> unit = "delete_bck_stub"
external delete_bck_stub : gcscion -> unit = "delete_bck_stub"
external set_root_distance : int -> unit = "set_root_distance"
external set_debug_flag : bool -> unit = "set_debug_flag"

let print_mark mark =
  Printf.printf "Mark: ";
  let ((a1,p1,_),(a2,p2,_),((a3,p3,_),id)) = mark.mark_id in
  Printf.printf "  Id: %s:%d %s:%d + %s:%d %d" 
    (string_of_inet_addr a1) p1
    (string_of_inet_addr a2) p2
    (string_of_inet_addr a3) p3
    id;
  print_newline ();
  Printf.printf "  Gen: %d" mark.mark_gen;
  print_newline () 
  

let rootmark = (Obj.magic 0: mark)

let root_distance = 
  try
    int_of_string (Sys.getenv "JCGC_RANGE")
  with
    _ -> 5

let _ = set_root_distance root_distance

let time = ref 0
let rootstamp = { stamp_mark = rootmark;
    stamp_color = GrayMark;
    stamp_distance = root_distance; }


let update_stub stub =
  let stamp = stub.gcstub_max in
  match stub.gcstub_bck_mark with
    None -> stamp
  | Some bck_mark ->
      let rec compute_color color bck_trace =
        match bck_trace with
          NoOtherScion -> color
        | GcScion scion ->
            if scion.gcscion_fwd_stamp.stamp_color = BlackMark ||
              scion.gcscion_fwd_stamp.stamp_mark <> bck_mark then
              BlackMark
            else
            if color <> WhiteMark then 
              compute_color color scion.gcscion_next
            else
            if scion.gcscion_fwd_stamp.stamp_color = WhiteMark ||
              scion.gcscion_fwd_stamp.stamp_mark = bck_mark then
              compute_color WhiteMark scion.gcscion_next
            else
              compute_color GrayMark scion.gcscion_next
      in
      if stamp.stamp_mark = bck_mark then
        { stamp_color = compute_color stamp.stamp_color stub.gcstub_bck_trace;
      stamp_mark = stamp.stamp_mark;
      stamp_distance = stamp.stamp_distance;
} else 
  stamp
          
(* after the local GC cycle (ie two local GC)*)

let gc_num = ref 0
let local_gc () = 
  incr gc_num;
  critical_section (function () ->
        let spaces = ref [] in
        incr time;
        let stubs = ref 0 in
        Hashtbl.iter 
          (fun addr space ->
            if not (space.space_id == Space.local_id) then
              spaces := space :: !spaces;
        ) Space.spaces;
        
        List.iter 
          (fun space ->
            
            let life_stubs = ref [] in
            for i = 0 to Weak.length space.stubs_table - 1 do
              match Weak.get space.stubs_table i with
                Some gcstub -> 
                  if not gcstub.gcstub_deleted then
                    life_stubs := (i, update_stub gcstub) :: !life_stubs
              | None -> ()
            done;
            
            let life_scions = ref [] in
            Array.iteri (fun i scion -> 
                match scion with
                  None -> ()
                | Some gcscion ->
                    life_scions := (i,gcscion.gcscion_deleted,gcscion.gcscion_bck_mark) :: !life_scions;
            
            ) space.scions_table;
            
            Debug.print1 'g' "Life stubs : %d\n" (List.length !life_stubs);
            Space.send space 
              (Dgc_msg (space.next_ts,
                Life (space.threshold, 
                  (if (!gc_num) mod 5 = 4 then
                      begin
                        (*Printf.printf "Cleaning the cache ...";
                        print_newline () ;*)
                        Jprims.cache_clean space.cache_out;
                        true
                      end
                    else
                      false),
                  !life_stubs,
                  !life_scions)));
            space.next_ts <- space.next_ts + 1;
            space.threshold <- space.current_ts;
            space.current_ts <- space.last_ts;
        ) !spaces;
        
        Whashtbl.clean Space.stubs_table)
    
(* Dgc control *)
let doit = ref true
let stop () = doit := false
let restart () = doit := true; local_gc ()
    
(* toplevel clean stop *)
let halted = ref false

let threads = ref 0
let init () =
  try
    Sys.getenv "JNODGC"; ()
  with
    Not_found ->
      Thread.create 
        (fun () ->
          while true do
            wait_on_gc ();
            Debug.print0 'g' "Major GC\n";
            if !doit then local_gc ()
          done
      ) ();
      incr threads

let rec remove_bck_stub bck_trace =
  match bck_trace with
    NoOtherScion -> ()
  | GcScion gcscion ->
      delete_bck_stub' gcscion;
      let bck_trace = gcscion.gcscion_next in
      gcscion.gcscion_next <- NoOtherScion;
      gcscion.gcscion_bck_mark <- None;
      remove_bck_stub bck_trace

let rec update_bck_trace bck_trace scion_bck_mark =
  match bck_trace with
    NoOtherScion -> NoOtherScion
  | GcScion gcscion ->
      let fwd_mark = gcscion.gcscion_fwd_stamp.stamp_mark in
      if fwd_mark.mark_id = scion_bck_mark.mark_id &&
        fwd_mark.mark_range = scion_bck_mark.mark_range then
(* this scion is OK for update *)
        begin
          gcscion.gcscion_bck_mark <- Some scion_bck_mark;
          gcscion.gcscion_next_stamp <-
            { stamp_color = WhiteMark;
            stamp_mark = scion_bck_mark;
            stamp_distance = scion_bck_mark.mark_range };
          gcscion.gcscion_next <-
            update_bck_trace gcscion.gcscion_next scion_bck_mark;
          gcscion
        end
      else
(* remove gcscion from bck_trace *)
        begin
          gcscion.gcscion_bck_mark <- None;
          let next = update_bck_trace gcscion.gcscion_next scion_bck_mark
          in
          gcscion.gcscion_next <- NoOtherScion;
          next
        end

let rec delete_bck_trace bck_trace =
  match bck_trace with
    NoOtherScion -> ()
  | GcScion gcscion -> 
      let bck_trace =  gcscion.gcscion_next in
      gcscion.gcscion_deleted <- true;
      gcscion.gcscion_next <- NoOtherScion;
      delete_bck_trace bck_trace

let update_stub_from_scion gcstub deleted bck_mark =
  if deleted then
    begin
      gcstub.gcstub_deleted <- true;
      delete_bck_trace gcstub.gcstub_bck_trace;
      gcstub.gcstub_bck_trace <- NoOtherScion
    end
  else
    begin
      match bck_mark, gcstub.gcstub_bck_mark with
        None,None -> ()
      | None,Some _ ->
(* sub-generation aborted *)
          remove_bck_stub gcstub.gcstub_bck_trace;
          gcstub.gcstub_bck_trace <- NoOtherScion;
          gcstub.gcstub_bck_mark <- None
      | Some _, None ->
(* sub-generation started *)
          gcstub.gcstub_bck_mark <- bck_mark
      | Some scion_bck_mark, Some stub_bck_mark ->
          if scion_bck_mark.mark_id = stub_bck_mark.mark_id &&
            scion_bck_mark.mark_range = stub_bck_mark.mark_range then
            begin
              if scion_bck_mark.mark_gen <> stub_bck_mark.mark_gen then
(* new sub-generation *)
                begin
                  gcstub.gcstub_bck_trace <- 
                    update_bck_trace gcstub.gcstub_bck_trace scion_bck_mark;
                  gcstub.gcstub_bck_mark <- bck_mark;
                end
            end
          else
(* old sub-generation aborted and new one started: could happen ??? *)
            begin
              remove_bck_stub gcstub.gcstub_bck_trace;
              gcstub.gcstub_bck_trace <- NoOtherScion;
              gcstub.gcstub_bck_mark <- bck_mark;
            end
    end

(* [update_scion_from_stub]: We have received a new [stamp] by a message
from the stub associated with that scion. We must update the value held
by the scion. *)
let update_scion_from_stub gcscion stamp =
  gcscion.gcscion_fwd_stamp <- stamp;
  gcscion.gcscion_next_stamp <-
    (match gcscion.gcscion_bck_mark with
      None ->
        
        if stamp.stamp_distance = 0 then
(* new generator *)
          let range = 
            Debug.print0 'g' "New generator\n";
            if stamp.stamp_mark = rootmark then root_distance
            else stamp.stamp_mark.mark_range
          in
          let mark = { mark_range = 2*range;
              mark_id = (Space.local_id, 
                gcscion.gcscion_space.space_id, 
                let locator = match 
                    gcscion.gcscion_target.locator with
                    None -> assert false | Some l -> l
                in locator.lctr_uid);
              mark_gen = !time; }
          in
          gcscion.gcscion_generator <- true;
          gcscion.gcscion_bck_mark <- Some mark;
          Debug.print0 'g' "G";
          { stamp_color = WhiteMark;
            stamp_distance = mark.mark_range;
            stamp_mark = mark }
        
        else
(* forwarder scion *)
          begin
(* we can mutate the stamp, since it is not shared 
(each stub has its own stamp) *)
            stamp.stamp_distance <- stamp.stamp_distance - 1;
            if stamp.stamp_mark == rootmark then
              Debug.print1 'g' "r%d" stamp.stamp_distance
            else
              begin
                Debug.print1 'g' "g%d" stamp.stamp_distance;
                print_mark stamp.stamp_mark;
              end;
            stamp
          end
    
    | Some bck_mark ->
        
        if gcscion.gcscion_generator then
          begin
            if (stamp.stamp_mark = rootmark ||
                stamp.stamp_mark > bck_mark) &&
              stamp.stamp_distance > 0
            then
(* generator erased by greater mark *)
              begin
                gcscion.gcscion_generator <- false;
                gcscion.gcscion_bck_mark <- None;
                stamp.stamp_distance <- stamp.stamp_distance - 1;
                if stamp.stamp_mark == rootmark then
                  Debug.print1 'g' "r%d" stamp.stamp_distance
                else
                  Debug.print1 'g' "g%d" stamp.stamp_distance;
                stamp
              end
            else
            if stamp.stamp_mark = bck_mark &&
              stamp.stamp_color = WhiteMark
            then
(* cycle detected *)
              begin
                Debug.print0 'g' "!<DELETED>";
                delete_bck_stub gcscion;
                gcscion.gcscion_deleted <- true;
                gcscion.gcscion_next_stamp
              end
            else
(* standard generator *)
              begin
                Debug.print0 'g' "g";
                gcscion.gcscion_next_stamp
              end
          end
        else
        if stamp.stamp_mark <> bck_mark then
(* remove scion from sub-generation *)
          begin
            delete_bck_stub gcscion;
            stamp.stamp_distance <- stamp.stamp_distance - 1;
            if stamp.stamp_mark == rootmark then
              Debug.print1 'g' "r%d" stamp.stamp_distance
            else
              Debug.print1 'g' "g%d" stamp.stamp_distance;
            stamp
          end
        else
(* standard sub-generator *)
          begin
            Debug.print0 'g' "b";
            { stamp_color = WhiteMark;
              stamp_mark = bck_mark;
              stamp_distance = bck_mark.mark_range;}
          end
  )

let inverse_assoc list =
  let rec iter old new_list =
    match old with
      (a,b)::tail -> iter tail ((b,a)::new_list)
    | [] -> new_list
  in
  iter list []

let tag_list = inverse_assoc Tag.list

let name_of_tag tag =
  try
    List.assoc tag tag_list
  with
    Not_found -> "block"


let life_handler space threshold life_stubs life_scions () =

(* update scions *)
  Debug.print0 'g' "Scions states : \n";
  List.iter (fun (i,fwd_stamp) -> 
      let gcscion =  match space.scions_table.(i) with
          None -> assert false | Some s -> s in
      gcscion.gcscion_marked <- true;
      let tag = Obj.tag (Obj.repr gcscion.gcscion_target) in
      Debug.print1 'g' "(%s)" (name_of_tag tag);
      if gcscion.gcscion_ts < threshold then
        (if not gcscion.gcscion_deleted then
            update_scion_from_stub gcscion fwd_stamp
          else
            Debug.print0 'g' "d";
        )
      else
        Debug.print0 'g' "v"
  ) life_stubs;
  Debug.print0 'g' "\n";

  let size = Array.length space.scions_table in
  let removed = ref 0 in
  for i = 0 to size - 1 do
    match space.scions_table.(i) with
      None -> ()
    | Some gcscion ->
        if gcscion.gcscion_marked || gcscion.gcscion_ts >= threshold then 
          gcscion.gcscion_marked <- false
        else
          begin
            let locator = match gcscion.gcscion_target.locator with
                Some l -> l | None -> assert false in
            Hashtbl.remove_binding space.scions_hash locator.lctr_uid i;
            space.free_list.(space.free_slots) <- i;
            space.scions_table.(i) <- None;
            space.free_slots <- space.free_slots + 1;
            incr removed;
            gcscion.gcscion_fwd_stamp.stamp_mark <- rootmark;
            delete_bck_stub gcscion;
            gcscion.gcscion_deleted <- true;
          end
  done;
  Debug.print1 'g' "Removed scions: %d\n" !removed;
  (* update stubs *)
  List.iter (fun (i,deleted,scion_bck_mark) ->
      if i < Weak.length space.stubs_table then
      match Weak.get space.stubs_table i with
        None -> ()
      | Some gcstub -> update_stub_from_scion gcstub deleted scion_bck_mark
  ) life_scions

let handler space msg = 
  match msg with
  | Life(threshold,clean,life_stubs,life_scions) ->
      if clean then Jprims.cache_clean space.cache_in;
      incr time;
      critical_section (life_handler space threshold life_stubs life_scions)
      
      
