(***********************************************************************)
(*                                                                     *)
(*                    JOCAML                                           *)
(*                                                                     *)
(*      Fabrice Le Fessant, projets PARA/SOR INRIA Rocquencourt        *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://pauillac.inria.fr/~lefessan            *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)


open Jtypes;;
open Reloc
open Dgc

  (* stub_locator: On extrait le locator (information sur la destination
  du stub) a partir du stub. Si le locator n'existe pas (le stub pointe
  sur un objet local non exporte' jusqu'a present), on cree le locator
  a la demande. 
  
  PS: On emploie le terme "stub" pour designer tout objet pouvant etre
  soit local, soit distant. Le stub est en effet une indirection toujours
  presente pour un tel objet.
  *)
  
let create_locator stub =
        (*  Note that objects which have never been exported may have
            a .gc_stub field incorrect (ie Val_unit instead of ref).
            *)
        Debug.print0 'v' "Create locator\n";
        let object_type =  match stub.local with
          RemoteObject -> 
            Debug.print0 'v' "For remote object\n";
            (* possible si migration *)
            let tag = Obj.tag (Obj.repr stub) in
            if tag = Tag.location then Location else
            if tag = Tag.automata then Automata else
            if tag = Tag.thread then Thread else
            if tag = Tag.object then 
              let class_info = Mop.get_class_info stub in
              Object (Space.local_id,
                      class_info.Mop.class_type,
                      class_info.Mop.class_sig,
	              class_info.Mop.table.Mop.size)
            else
              assert false
            
        | FailedObject object_type -> object_type
        | LocalThread thread -> Thread
        | LocalAutomata automata -> Automata
        | LocalLocation location -> 
            if Obj.tag (Obj.repr stub) = Tag.location then Location
            else
              let class_info = Mop.get_class_info stub in
              Object (Space.local_id,
                      class_info.Mop.class_type,
                      class_info.Mop.class_sig,
	              class_info.Mop.table.Mop.size)
	in
        let uid = Space.new_uid () in
        let gc_stub = ref LocalStub in
        stub.gc_stub <- gc_stub;
        let locator =
          { lctr_age = 0;
            lctr_weakspace = Space.local_id;
            lctr_weaknum = Whashtbl.add_direct Space.stubs_table uid (Some stub);
            lctr_uid = uid;
            lctr_type = object_type;
          } in
        stub.locator <- Some locator;
        locator

      (* create_scion: on cree un nouveau scion dans le stub qui sera gere
      par le GC distribue. *)
      
let create_scion dest_space ts stub locator =  
  let scion_num =
    if dest_space.free_slots = 0 then
      begin
        let old_size = Array.length dest_space.scions_table in
        let new_size = 2*old_size+10 in
        let new_table = Array.init new_size 
            (function i ->
                if i<old_size then dest_space.scions_table.(i)
                else None)
        in
        let new_list = Array.init new_size
            (function i -> new_size - i - 1) in
        dest_space.free_slots <- new_size - old_size;
        dest_space.scions_table <- new_table;
        dest_space.free_list <- new_list;
      end;
    dest_space.free_slots <- dest_space.free_slots -1;
    dest_space.free_list.(dest_space.free_slots)
  in 
  let gc_scion =
    {[gcscion]
        gcscion_stamp = rootstamp;
      gcscion_next_stamp = rootstamp;
      gcscion_bck_mark = None;
      gcscion_bck_stub = NoBackTrace;
      gcscion_generator =  false;
      gcscion_deleted = false;
      gcscion_fwd_stamp = rootstamp;
      gcscion_next = NoOtherScion;
      gcscion_ts = ts;
      gcscion_space = dest_space;
      gcscion_num = scion_num;
      gcscion_age = locator.lctr_age;
      gcscion_stub = stub.gc_stub;
      gcscion_target = stub;
      gcscion_marked = false;
    } in
  dest_space.scions_table.(scion_num) <- Some gc_scion;
  Hashtbl.add dest_space.scions_hash locator.lctr_uid scion_num;
  scion_num

  (* update_scion: si le scion existe quand on exporte une reference, on 
  met a jour le timestamp du scion. *)
  
let update_scion dest_space ts stub locator scion_num =  
  let gcscion = match dest_space.scions_table.(scion_num) with
      None -> assert false | Some s -> s in
  if gcscion.gcscion_age <> locator.lctr_age then raise Not_found;
  if gcscion.gcscion_ts < ts then gcscion.gcscion_ts <- ts;
  
  if gcscion.gcscion_generator then
    begin
      gcscion.gcscion_bck_mark <- None;
      gcscion.gcscion_generator <- false;
    end;
  gcscion.gcscion_next_stamp <- rootstamp;
  gcscion.gcscion_fwd_stamp <- rootstamp

  (* ptr2loc: fournit une reference speciale pour les objets pouvant
  etre exportes par une reference distante. *)
  
let rec ptr2loc dest_space ts (stub:generic_stub) =  
  Debug.print0 'v' "ptr2loc:\n";
  let locator = match stub.locator with
      None -> create_locator stub 
    | Some locator -> 
        Debug.print0 'v' "Deja un locator\n";
        locator in  
  let scion_num =   
    try
      let scion_num = Hashtbl.find dest_space.scions_hash locator.lctr_uid in
      Debug.print0 'v' "Deja un GC scion\n";
      update_scion dest_space ts stub locator scion_num;
      scion_num
    with    
      Not_found ->
        Debug.print0 'v' "On cree un nouveau GC scion\n";
        create_scion dest_space ts stub locator
  in
  StubLocator (scion_num,locator)

  (* ptr2obj: fournit une reference speciale pour les objets ne pouvant
  etre exportes par une reference normale. *)


let ptr2obj dest_space ts pointer =
  let tag = Obj.tag pointer in  
  if tag = Tag.reloc then
    let module_data = (Obj.magic pointer : module_data) in
    if module_data.migrate && 
      not(List.mem module_data.module_name dest_space.modules) then
      begin
        dest_space.modules <- module_data.Reloc.module_name
          :: dest_space.modules;
        RelocModule module_data
      end
    else
      DummyModule (module_data.module_name,Obj.size pointer)
  else  
  if tag = Tag.exn then
    let exn = (Obj.magic pointer : exn) 
    in
    RelocException(exn.exn_name,exn.mod_name,exn.mod_num)
  else
  if tag = Tag.marshal then
    SpecialMarshal(Obj.magic (Obj.field pointer 0),
      (Obj.magic (Obj.field pointer 1)) pointer)
  else
  if tag = Tag.classinfo then
    let cl = (Obj.magic pointer: Mop.class_info) in
    Class_info cl.Mop.class_type
  else
    ptr2loc dest_space ts (Obj.magic pointer : generic_stub)
    
    (* marshal: transforme un message contenant des donnees Join (modules,
    references distantes, exception,etc) en un message portant ces informations
    sous forme symbolique. Le message est d'abord passe au marshaler 
    d'Ocaml, qui retourne une liste des objets qu'il n'a pu emballer, ainsi
    qu'une chaine contenant des "trous" pour chaque objet de la liste.
    La liste est ensuite traitee pour transformer chaque reference en une
    donnee symbolique. L'ensemble peut alors etre repasser dans le marshaler
    sans generer d'erreur. *)
    
let marshal dest_space message =
  
  let message_name =
    match message with
      GetService _ -> "GETSERVICE"
    | Synchrone _ -> "SYNCHRONE"
    | Invoke _ -> "INVOKE"
    | Asynchrone _ -> "ASYNCHRONE"
    | LocationMigration _ -> "LOCATIONMIGRATION"
    | ReplyThread _ -> "REPLYTHREAD"
    | UpdateStub _ -> "UPDATESTUB"
  in
  Debug.print1 's' "Marshal: %s\n" message_name;
  
  let ts = dest_space.next_ts in
  dest_space.next_ts <- dest_space.next_ts + 1;
  let (buffer, pointers) = 
    Jprims.marshal_message message dest_space.cache_out in
  Debug.print1 's' "Sending %d locators\n" (Array.length pointers);
  let locators = Array.map (ptr2obj dest_space ts) pointers in
  Debug.print2 'z' "Marshal %s (%d)\n" message_name ts;
  Join_msg (ts, buffer, locators)
;;

let send dest_space message =
  Debug.print "Locator.send";
  Space.send dest_space (marshal dest_space message)
;;

  
