(***********************************************************************)
(*                                                                     *)
(*                    Join : join.ml                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;

type classe

external filter_channel : (('a -> 'b) as 'c) -> 'c = "%identity"

let go = Go.go;;

let kill location =
  if location = Space.local_location then
    Pervasives.exit 0 
  else
    Jprims.kill_location location

let halt () = 
  Debug.print0 'g' "Join.halt...";
  let location = Jprims.self_location () in
  kill location

type location = Jtypes.location_stub

(* Join-calculus *)

let bar process1 process2 =
  Thread.safe_create process1 (); process2 ()
;;

let launch process = Thread.safe_create process (); ();;

let server () = Jprims.server ()
let here = Jprims.self_location ()
let self_location () = Jprims.self_location ()

let go = Go.go

let _ =
  Space.join_handler := Handler.handler;
  Space.join_initialised := true;
();; (* on autorise les connections *)
  


(* *)

(* L'interface *)
type space = Jtypes.space
type process = Jtypes.process;;
type automata = Jtypes.automata_stub;;
type channel = Jtypes.channel;;
type module_data = Reloc.module_data;;
type thread_stub = Jtypes.thread_stub;;
type reply_event = thread_stub
type space_id = Jtypes.space_id


let create_empty_automata = Jprims.create_empty_automata
let fill_automata = Jprims.fill_def;;

  
let reply_to th v = 
  Channel.reply_to th (Reply (Obj.repr v))

let reply_to_and_exit th v = 
  Channel.reply_to_and_exit th (Reply (Obj.repr v))

let send_exception exn thread_stubs = Channel.send_exception exn thread_stubs

let space_of_addr = Space.space_of_addr
let space_of_string name port = 
  space_of_addr (Space.first_addr (Unix.gethostbyname name)) port

let getRemoteService = Services.getRemoteService;;
let setLocalService = Services.setLocalService;;
let getLocalService = Services.getLocalService;;

let create_agent module_data modules =
  let self_loc = Jprims.self_location () in
  let location =
    Jprims.create_new_location Space.local_space self_loc module_data 
  in
  match location.local with
    LocalLocation local ->
      local.modules <- modules; location
  | _ -> assert false

let agent_thread location f = Jprims.thread_create_in location f ()

let exit () = Thread.exit ()

let safe_tail_synchrone = Channel.safe_tail_synchrone
let unsafe_tail_synchrone = Channel.unsafe_tail_synchrone

let reuse_thread thread_stub clos = Jprims.reuse_thread clos () thread_stub

let print i =
  print_int i;
  flush stdout

let definition () =
  let (automata,defun,declare) = Jprims.create_empty_automata
    (Jprims.self_location ())
    0 2 [|1;0|] [|true;false|]
  in
  let process = function 
      (x,(reply_defun,())) ->
        reply_to_and_exit (* join.mlg 17 *) reply_defun x
  in
  fill_automata automata [|Obj.magic process|] (0,2,(3,0),(3,0),((0,2),0,1));
  ((declare,defun): (<<'a>>) * (unit -> 'a))

let _ = Dgc.init ()
let dgc_stop = Dgc.stop
let dgc_restart = Dgc.restart

external test_channel' : 's -> string -> unit = "test_good_channel"
let test_channel c s = test_channel' s c

let goo oo = go (Obj.magic oo: location)
let fail obj = 
  let stub = Jprims.extract_stub obj in
  Jprims.wait_on_fail stub

let new_port = Space.new_port
let migrate oo dest = Jprims.migrate (Obj.magic oo,dest)
let migrate_to self dest = (Obj.magic self)#migrate_to dest

let spawn_send (channel : << 'a >>) (arg : 'a) =  (Obj.magic channel) arg

external check: 'a -> int -> unit = "check_pointer"

open Mop
  
let create_automata_class meths module_data =
  let class_info = Obj.new_block Tag.classinfo 5 in
  let (class_info : Mop.class_info) = Obj.magic class_info in
  let pub_meths = List.map (fun (meth,_,_) -> meth) meths in
  let cl_init table =
    List.iter (fun (met, chan, sync) ->
        Mop.set_method table met 
        1 
          (Obj.magic ((if sync then Channel.send_synchrone_obj
                else Channel.send_asynchrone_obj) chan): Mop.item)
    ) meths
  in
  let obj_creator = Obj.magic (fun table -> ()) in
  Mop.create_class class_info pub_meths obj_creator cl_init module_data;
  class_info.table.ext_buckets <- class_info.table.buckets;
  class_info
    
  
