(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

(*
  Tout stub pouvant etre reference depuis l'exterieur possede au
  moins un gc_stub. Si l'objet est local, le gc_stub pointe vers
  l'espace courant.
*)

open Jtypes
open Reloc
open Dgc

let create_stub src_space ts locator =
  let stub =
    match locator.lctr_type with
      Object (space_id,class_id,class_sig,obj_size) ->
        let stub = Obj.new_block Tag.object obj_size in
        let space = Space.get_space space_id in
        Obj.set_field stub 5 
          (Obj.repr
            (Channel.remote_ext_buckets space class_id class_sig));
        stub
    | Automata -> Obj.new_block Tag.automata 5
    | Thread ->   Obj.new_block Tag.thread 5
    | Location -> 
        Obj.new_block Tag.location 5
  in
  let stub = (Obj.magic stub : generic_stub) in
  Whashtbl.add Space.stubs_table locator.lctr_uid (Some stub);
  stub.locator <- Some locator;
  let space =  Space.get_space locator.lctr_weakspace in
  stub.position <- space;
  locator.lctr_weakspace <- space.space_id;
  stub.local <- 
    (if stub.position == Space.local_space then  
      FailedObject locator.lctr_type
    else 
      RemoteObject);
        (* These stubs have incorrect fields .gc_stub *)
  stub.gc_stub <- ref LocalStub;
  stub

    
let update_stub src_space stub locator =
  let old_locator = match stub.locator with
      None -> assert false 
    | Some l -> l in
  if old_locator.lctr_age < locator.lctr_age then
      (* the object has moved since creation of that locator *)
    let new_space = Space.get_space locator.lctr_weakspace in
    stub.position <- new_space;
    old_locator.lctr_age <- locator.lctr_age;
    old_locator.lctr_weaknum <- locator.lctr_weaknum;
    old_locator.lctr_weakspace <- new_space.space_id;
    
    stub.gc_stub <- ref LocalStub;
    stub.cache <- 0;
    if new_space <> Space.local_space then
      stub.local <- RemoteObject

let create_gcstub src_space stub locator scion_num =
(* LocalStub: new stub, or migrating one
   IndirectStub: the new one is Direct
*)
  if scion_num >= Weak.length src_space.stubs_table then
    begin
      let old_size = Weak.length src_space.stubs_table in
      let new_size = 2 * old_size + 10 in
      let new_table = Weak.create new_size in
      Weak.blit src_space.stubs_table 0 new_table 0 old_size;
      src_space.stubs_table <- new_table;
    end;
  let gcstub = 
    match Weak.get src_space.stubs_table scion_num with
      Some gcstub -> gcstub
    | None ->
              (* We must create a new gc stub *)
        let gcstub = 
          {[gcstub]     
              gcstub_min = { stamp_mark = rootmark; 
              stamp_color = GrayMark; 
              stamp_distance = root_distance };
            gcstub_max = { stamp_mark = rootmark; 
              stamp_color = GrayMark; 
              stamp_distance = root_distance };
            gcstub_bck_mark = None;
            gcstub_bck_trace = NoOtherScion;
            gcstub_obj = stub;
            gcstub_space = src_space;
            gcstub_age = locator.lctr_age; 
            gcstub_num = scion_num;
            gcstub_deleted = false;
          }
        in
        Weak.set src_space.stubs_table scion_num (Some gcstub);
        gcstub
  in
  
  
  if stub.position == src_space then (* direct *)
    stub.gc_stub := DirectStub gcstub
  else
(* here, we could keep all indirections when the cycles
   detector will be implemented, and proved efficient *)
    stub.gc_stub := IndirectStub gcstub  
      
let rec loc2ptr src_space ts scion_num locator =

  let stub =
    try
      
      let stub =
        if locator.lctr_weakspace = Space.local_id then 
          Space.find_stub locator.lctr_weaknum locator.lctr_uid
        else
          match
            Whashtbl.find Space.stubs_table locator.lctr_uid 
          with
            None -> 
              Whashtbl.remove Space.stubs_table locator.lctr_uid;
              raise Not_found
          | Some stub -> stub
      in
      update_stub src_space stub locator;
      stub

    with
      Not_found ->
        create_stub src_space ts locator

  in
  if stub.position == Space.local_space then stub else
  match !(stub.gc_stub) with
    DirectStub gcstub -> 
(* we already have a direct gcstub *)
      stub

  | IndirectStub gcstub when not (stub.position == src_space)->
(* Indirect but the new one also *)
      stub
  | _ ->
      create_gcstub src_space stub locator scion_num;
      stub


exception MessageRefused;;

(* Il faut ajouter ces informations dans Space. *)
let unmarshalers = ref []
let new_unmarshaler (name:string) (f: 'a -> 'b) = 
  unmarshalers := (name,Obj.magic f) :: !unmarshalers
let lookup name = List.assoc name !unmarshalers 
  
let unmarshal (space : space) ts buffer locators =
  if ts < space.last_ts then raise MessageRefused;
  space.last_ts <- ts;
  let pointers =
    Array.map 
      (function 
        DummyModule (name,size) ->
          Obj.repr (find_module_data name size )
      | RelocModule module_data -> 
          let m = find_module_data module_data.module_name
              (Obj.size (Obj.repr module_data))
          in
          install_module module_data m;
          Obj.repr m
      | RelocException (exn_name,mod_name,mod_num) ->
          let e = find_exn exn_name mod_name mod_num in
          Obj.repr e
      | StubLocator(scion_num, locator) -> 
          Obj.repr (loc2ptr space ts scion_num locator)
      | SpecialMarshal(name,t) -> (lookup name) t
      | Class_info cl ->
          let class_info = 
            try
              Hashtbl.find Mop.classes cl
            with _ ->
                let class_info = Obj.new_block Tag.classinfo 5 in
                let class_info = (Obj.magic class_info: Mop.class_info) in
                Hashtbl.add Mop.classes cl class_info;
                class_info
          in
          Obj.repr class_info
    ) locators
  in
  Jprims.unmarshal_message buffer pointers space.cache_in

