(***********************************************************************)
(*                                                                     *)
(*                    Join : jutils.ml                                 *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*************************----------------------------            Join *)

let go = Join.go
let kill = Join.kill
let halt = Join.halt
let top_location = Join.here
let getRemoteService = Join.getRemoteService
let setLocalService = Join.setLocalService
let getLocalService = Join.getLocalService
let space_of_addr = Join.space_of_addr
let space_of_string = Join.space_of_string
let dgc_stop = Join.dgc_stop
let dgc_restart = Join.dgc_restart
let goo = Join.goo
let fail = Join.fail
let new_port = Join.new_port
let migrate = Join.migrate
let migrate_to = Join.migrate_to

(*************************----------------------------             I/O *)

type t
external open_descriptor_out: int -> out_channel = "caml_open_descriptor"
external open_descriptor_in: int -> in_channel = "caml_open_descriptor"

type out_channel = {
    flush : (unit -> unit);
    output_char : (char -> unit);
    output_string : (string -> unit);
    output_byte : (int -> unit);
    output_binary_int : (int -> unit);
    output_value : (t -> unit);
    seek_out : (int -> unit);
    pos_out : (unit -> int);
    out_channel_length : (unit -> int);
    close_out : (unit -> unit);
    local_from_out_channel : (unit -> local_out_channel);
  } 

type in_channel = 
    {
    input_char : (unit -> char);
    input_line : (unit -> string);
    input : (int -> string);
    really_input : (int -> string);
    input_byte : (unit -> int);
    input_binary_int : (unit -> int);
    input_value : (unit -> t);
    seek_in : (int -> unit);
    pos_in : (unit -> int);
    in_channel_length : (unit -> int);
    close_in : (unit -> unit);
    unsafe_really_input : (int -> string);
    local_from_in_channel : (unit -> local_in_channel);
  }
            
(* wrappers *)
type rep =
    InChannel of in_channel
  | OutChannel of out_channel
  | Exception of exn

external repr : 'a -> t = "%identity"
external magic : t -> 'a = "%identity"


let out_open f =
  let def wait_rep () | rep! r = reply r in
  let loc location 
    do { Join.go Join.here; 
         (try
           let outc = f () in
           let def flush' () = flush outc; reply
             or output_char' c = output_char outc c; reply
             or output_string' s = output_string outc s; reply
             or output_byte' i = output_byte outc i; reply
             or output_binary_int' i = output_binary_int outc i; reply
             or output_value' t = output_value outc t; reply
             or close_out' () =  close_out outc; reply
             or seek_out' i = seek_out outc i;reply
             or pos_out' () = reply (pos_out outc)
             or out_channel_length' () = reply (out_channel_length outc)
           in
           let chan = {
             flush = flush';
             output_char = output_char';
             output_string = output_string';
             output_byte = output_byte';
             output_binary_int = output_binary_int';
             output_value = output_value';
             seek_out = seek_out';
             pos_out = pos_out';
             out_channel_length = out_channel_length';
             close_out = close_out';
             local_from_out_channel = 
             (function () -> failwith "Channel is not local");
           } 
           in
           {| rep (OutChannel chan) }
         with
           e -> {| rep (Exception e)}
               ); }
  in
  match   wait_rep () with
    OutChannel chan ->  (Obj.magic (chan : out_channel) : Pervasives.out_channel)
  | Exception e -> raise e
  | _ -> assert false

let in_open f =
  let def wait_rep () | rep! r = reply r in
  let loc location 
    do { Join.go Join.here; 
         (try
           let inc = f () in
           let def input_char'  () = reply (input_char inc)
             or input_line' () = reply (input_line inc)
             or input' len =  
                let buf = String.create len in
                let len = input inc buf 0 len in
                reply (String.sub buf 0 len)
             or really_input' len =
                let buf = String.create len in
                really_input inc buf 0 len;
                reply buf
             or input_byte' () = reply (input_byte inc)
             or input_binary_int' () = reply (input_binary_int inc)
             or input_value' ()= reply (input_value inc)
             or seek_in' i = seek_in inc i; reply
             or pos_in' () = reply (pos_in inc)
             or in_channel_length' () = reply (in_channel_length inc)
             or close_in' () = close_in inc; reply
             or unsafe_really_input' len =
                let buf = String.create len in
                unsafe_really_input inc buf 0 len;
                reply buf
           in
           let chan = {
             input_char = input_char';
             input_line = input_line';
             input = input';
             really_input = really_input';
             input_byte = input_byte';
             input_binary_int = input_binary_int';
             input_value = input_value';
             seek_in = seek_in';
             pos_in = pos_in';
             in_channel_length = in_channel_length';
             close_in = close_in';
             unsafe_really_input = unsafe_really_input';
             local_from_in_channel = 
             (function () -> failwith "Channel is not local");
           } 
           in
           {| rep (InChannel chan) }
         with
           e -> {| rep (Exception e)}
               ); }
  in
  match   wait_rep () with
    InChannel chan -> (Obj.magic (chan : in_channel) : Pervasives.in_channel)
  | Exception e -> raise e
  | _ -> assert false


let jopen_out filename = out_open (function () -> open_out filename)
let jopen_out_bin filename = out_open (function () -> open_out_bin filename)
let jopen_out_gen mode rights filename =
  out_open (function () -> open_out_gen mode rights filename)

let jopen_in filename = in_open (function () -> open_in filename)
let jopen_in_bin filename = in_open (function () -> open_in_bin filename)
let jopen_in_gen mode rights filename = 
    in_open (function () -> open_in_gen mode rights filename)

let remote_from_in_channel inc =   in_open (function () -> inc)
let remote_from_out_channel outc =  out_open (function () -> outc)
